package sun.misc;

import java.util.Arrays;

public class FormattedFloatingDecimal
{
    private int decExponentRounded;
    private char[] mantissa;
    private char[] exponent;
    private static final ThreadLocal<Object> threadLocalCharBuffer;
    
    public static FormattedFloatingDecimal valueOf(double n, int n2, FormattedFloatingDecimal.Form form) {
        return new FormattedFloatingDecimal(n2, form, FloatingDecimal.getBinaryToASCIIConverter(n, form == FormattedFloatingDecimal.Form.COMPATIBLE));
    }
    
    private static char[] getBuffer() {
        return (char[])(char[])FormattedFloatingDecimal.threadLocalCharBuffer.get();
    }
    
    private FormattedFloatingDecimal(int n, FormattedFloatingDecimal.Form form, FloatingDecimal.BinaryToASCIIConverter binaryToASCIIConverter) {
        if (binaryToASCIIConverter.isExceptional()) {
            this.mantissa = binaryToASCIIConverter.toJavaFormatString().toCharArray();
            this.exponent = null;
            return;
        }
        char[] buffer = getBuffer();
        int digits = binaryToASCIIConverter.getDigits(buffer);
        int decimalExponent = binaryToASCIIConverter.getDecimalExponent();
        int negative = binaryToASCIIConverter.isNegative();
        switch (form) {
            case COMPATIBLE: {
                this.fillCompatible(n, buffer, digits, this.decExponentRounded = decimalExponent, (boolean)negative);
                break;
            }
            case DECIMAL_FLOAT: {
                int applyPrecision = applyPrecision(decimalExponent, buffer, digits, decimalExponent + n);
                this.fillDecimal(n, buffer, digits, applyPrecision, (boolean)negative);
                this.decExponentRounded = applyPrecision;
                break;
            }
            case SCIENTIFIC: {
                int applyPrecision2 = applyPrecision(decimalExponent, buffer, digits, n + 1);
                this.fillScientific(n, buffer, digits, applyPrecision2, (boolean)negative);
                this.decExponentRounded = applyPrecision2;
                break;
            }
            case GENERAL: {
                int applyPrecision3 = applyPrecision(decimalExponent, buffer, digits, n);
                if (applyPrecision3 - 1 < -4 || applyPrecision3 - 1 >= n) {
                    --n;
                    this.fillScientific(n, buffer, digits, applyPrecision3, (boolean)negative);
                }
                else {
                    n -= applyPrecision3;
                    this.fillDecimal(n, buffer, digits, applyPrecision3, (boolean)negative);
                }
                this.decExponentRounded = applyPrecision3;
                break;
            }
            default: {
                assert false;
                break;
            }
        }
        return;
    }
    
    public int getExponentRounded() {
        return this.decExponentRounded - 1;
    }
    
    public char[] getMantissa() {
        return this.mantissa;
    }
    
    public char[] getExponent() {
        return this.exponent;
    }
    
    private static int applyPrecision(int n, char[] a, int toIndex, int fromIndex) {
        if (fromIndex >= toIndex || fromIndex < 0) {
            return n;
        }
        if (fromIndex == 0) {
            if (a[0] >= '5') {
                a[0] = '1';
                Arrays.fill(a, 1, toIndex, '0');
                return n + 1;
            }
            Arrays.fill(a, 0, toIndex, '0');
            return n;
        }
        else {
            if (a[fromIndex] >= '5') {
                int n2 = a[--fromIndex];
                if (n2 == '9') {
                    while (n2 == '9' && fromIndex > 0) {
                        n2 = a[--fromIndex];
                    }
                    if (n2 == '9') {
                        a[0] = '1';
                        Arrays.fill(a, 1, toIndex, '0');
                        return n + 1;
                    }
                }
                a[fromIndex] = (char)(n2 + 1);
                Arrays.fill(a, fromIndex + 1, toIndex, '0');
            }
            else {
                Arrays.fill(a, fromIndex, toIndex, '0');
            }
            return n;
        }
    }
    
    private void fillCompatible(int n, char[] array, int a, int n2, boolean b) {
        int n3 = b;
        if (n2 > 0 && n2 < 8) {
            if (a < n2) {
                int n8 = n2 - a;
                System.arraycopy(array, 0, this.mantissa = create(b, a + n8 + 2), n3, a);
                Arrays.fill(this.mantissa, n3 + a, n3 + a + n8, '0');
                this.mantissa[n3 + a + n8] = '.';
                this.mantissa[n3 + a + n8 + 1] = '0';
            }
            else if (n2 < a) {
                int min = Math.min(a - n2, n);
                System.arraycopy(array, 0, this.mantissa = create(b, n2 + 1 + min), n3, n2);
                this.mantissa[n3 + n2] = '.';
                System.arraycopy(array, n2, this.mantissa, n3 + n2 + 1, min);
            }
            else {
                System.arraycopy(array, 0, this.mantissa = create(b, a + 2), n3, a);
                this.mantissa[n3 + a] = '.';
                this.mantissa[n3 + a + 1] = '0';
            }
        }
        else if (n2 <= 0 && n2 > -3) {
            int max = Math.max(0, Math.min(-n2, n));
            int max2 = Math.max(0, Math.min(a, n + n2));
            if (max > 0) {
                (this.mantissa = create(b, max + 2 + max2))[n3] = '0';
                this.mantissa[n3 + 1] = '.';
                Arrays.fill(this.mantissa, n3 + 2, n3 + 2 + max, '0');
                if (max2 > 0) {
                    System.arraycopy(array, 0, this.mantissa, n3 + 2 + max, max2);
                }
            }
            else if (max2 > 0) {
                (this.mantissa = create(b, max + 2 + max2))[n3] = '0';
                this.mantissa[n3 + 1] = '.';
                System.arraycopy(array, 0, this.mantissa, n3 + 2, max2);
            }
            else {
                (this.mantissa = create(b, 1))[n3] = '0';
            }
        }
        else {
            if (a > 1) {
                (this.mantissa = create(b, a + 1))[n3] = array[0];
                this.mantissa[n3 + 1] = '.';
                System.arraycopy(array, 1, this.mantissa, n3 + 2, a - 1);
            }
            else {
                (this.mantissa = create(b, 3))[n3] = array[0];
                this.mantissa[n3 + 1] = '.';
                this.mantissa[n3 + 2] = '0';
            }
            int n4 = n2 <= 0;
            int n5;
            int n6;
            if (!(n4 == 0)) {
                n5 = -n2 + 1;
                n6 = 1;
            }
            else {
                n5 = n2 - 1;
                n6 = 0;
            }
            if (n5 <= 9) {
                (this.exponent = create((boolean)n4, 1))[n6] = (char)(n5 + 48);
            }
            else if (n5 <= 99) {
                (this.exponent = create((boolean)n4, 2))[n6] = (char)(n5 / 10 + 48);
                this.exponent[n6 + 1] = (char)(n5 % 10 + 48);
            }
            else {
                (this.exponent = create((boolean)n4, 3))[n6] = (char)(n5 / 100 + 48);
                int n7 = n5 % 100;
                this.exponent[n6 + 1] = (char)(n7 / 10 + 48);
                this.exponent[n6 + 2] = (char)(n7 % 10 + 48);
            }
        }
    }
    
    private static char[] create(boolean b, int n) {
        if (b) {
            char[] array = new char[n + 1];
            array[0] = '-';
            return array;
        }
        return new char[n];
    }
    
    private void fillDecimal(int n, char[] array, int a, int n2, boolean b) {
        int n3 = b;
        if (n2 > 0) {
            if (a < n2) {
                System.arraycopy(array, 0, this.mantissa = create(b, n2), n3, a);
                Arrays.fill(this.mantissa, n3 + a, n3 + n2, '0');
            }
            else {
                int min = Math.min(a - n2, n);
                System.arraycopy(array, 0, this.mantissa = create(b, n2 + ((min > 0) ? (min + 1) : 0)), n3, n2);
                if (min > 0) {
                    this.mantissa[n3 + n2] = '.';
                    System.arraycopy(array, n2, this.mantissa, n3 + n2 + 1, min);
                }
            }
        }
        else if (n2 <= 0) {
            int max = Math.max(0, Math.min(-n2, n));
            int max2 = Math.max(0, Math.min(a, n + n2));
            if (max > 0) {
                (this.mantissa = create(b, max + 2 + max2))[n3] = '0';
                this.mantissa[n3 + 1] = '.';
                Arrays.fill(this.mantissa, n3 + 2, n3 + 2 + max, '0');
                if (max2 > 0) {
                    System.arraycopy(array, 0, this.mantissa, n3 + 2 + max, max2);
                }
            }
            else if (max2 > 0) {
                (this.mantissa = create(b, max + 2 + max2))[n3] = '0';
                this.mantissa[n3 + 1] = '.';
                System.arraycopy(array, 0, this.mantissa, n3 + 2, max2);
            }
            else {
                (this.mantissa = create(b, 1))[n3] = '0';
            }
        }
    }
    
    private void fillScientific(int b, char[] array, int n, int n2, boolean b2) {
        int n3 = b2;
        int max = Math.max(0, Math.min(n - 1, b));
        if (max > 0) {
            (this.mantissa = create(b2, max + 2))[n3] = array[0];
            this.mantissa[n3 + 1] = '.';
            System.arraycopy(array, 1, this.mantissa, n3 + 2, max);
        }
        else {
            (this.mantissa = create(b2, 1))[n3] = array[0];
        }
        int n4;
        int n5;
        if (n2 <= 0) {
            n4 = '-';
            n5 = -n2 + 1;
        }
        else {
            n4 = '+';
            n5 = n2 - 1;
        }
        if (n5 <= 9) {
            this.exponent = new char[] { n4, '0', (char)(n5 + 48) };
        }
        else if (n5 <= 99) {
            this.exponent = new char[] { n4, (char)(n5 / 10 + 48), (char)(n5 % 10 + 48) };
        }
        else {
            int n6 = (char)(n5 / 100 + 48);
            int n7 = n5 % 100;
            this.exponent = new char[] { n4, n6, (char)(n7 / 10 + 48), (char)(n7 % 10 + 48) };
        }
    }
    
    static {
        threadLocalCharBuffer = new ThreadLocal<Object>() {
            @Override
            protected Object initialValue() {
                return new char[20];
            }
        };
    }
    
    public enum Form
    {
        SCIENTIFIC, 
        COMPATIBLE, 
        DECIMAL_FLOAT, 
        GENERAL;
    }
}
